/*
 * socketlock.c
 * -
 * Copyright (c) 2003 Luis Grangeia 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dlfcn.h>

/* Maximum number of exceptions to bind() before connect(): */
#define MAX_EXCEPTIONS 32

typedef struct {
	struct in_addr addr;
	struct in_addr mask;
} network_t;

static void *connect_p, *dl_handle, *bind_p, *accept_p;
static int initialized = 0;
static struct in_addr bind_addr;

static network_t except_nets[MAX_EXCEPTIONS];

int initialize();
void add_exceptions(network_t *, char *);
int bind(int, const struct sockaddr *, socklen_t);
int connect(int, const struct sockaddr *, socklen_t);


/* TODO: use dl's _init function */
int initialize() {
	char *str;
	int i;

#ifdef SUNOS
	dl_handle = dlopen("libsocket.so.1", RTLD_LAZY);
#else
	dl_handle = dlopen("libc.so.6", RTLD_LAZY);
#endif
	if (!dl_handle) {
		fprintf (stderr, "socketlock: error in dlopen(): %s\n", dlerror());
		return 0;
	}
	
	bind_p = dlsym(dl_handle, "bind");
	if (!bind_p) {
		fprintf (stderr, "socketlock: error in dlsym(): %s\n", dlerror());
		return 0;
	}

        connect_p = dlsym(dl_handle, "connect");
	if (!connect_p) {
		fprintf (stderr, "socketlock: error in dlsym(): %s\n", dlerror());
		return 0;
	}

	str = getenv("BINDTO");
	if (!str) {
		fprintf (stderr, "socketlock: error, please set BINDTO variable\n");
		return 0;
	}
	if(!inet_aton(str, &bind_addr)) {
		fprintf (stderr, "socketlock: %s is not a valid IP address\n", str);
		return 0;
	}
	
	str = getenv("EXEMPT_BIND");
	add_exceptions(except_nets, str);

#ifdef DEBUG
	printf("socketlock: will not bind connections to addresses/masks:\n");
	for(i=0; (except_nets[i].addr.s_addr && i < MAX_EXCEPTIONS); i++)
		printf("address: %x, mask: %x\n", except_nets[i].addr.s_addr, except_nets[i].mask.s_addr);
#endif
	printf("libsocketlock.so loaded ok\n");

	initialized = 1;
	return 1;
}


void add_exceptions(network_t ptr[], char *addrs_str) {
	int i, x;
	char *addr_pair_str, *addr_str, *mask_str, *t;
	static struct in_addr temp_addr;
	
	bzero(ptr, sizeof(network_t)*MAX_EXCEPTIONS);

	inet_aton("127.0.0.0", &ptr[0].addr);
	inet_aton("255.0.0.0", &ptr[0].mask);

	for (i = 1; (addrs_str && i < MAX_EXCEPTIONS); i++) { 
		addr_pair_str = addrs_str;

		t = strchr(addrs_str, ':');
		if (t) {
			*t = '\0';
			addrs_str = ++t;
		} else
			addrs_str = NULL;

		if(!addr_pair_str) continue;

		addr_str = addr_pair_str;
		t = strchr(addr_pair_str, '/');

		if (t) {
			*t = '\0';
			mask_str = ++t;
		} else
			mask_str = NULL;
			
		if (!addr_str) continue;
		if (!inet_aton(addr_str, &ptr[i].addr)) continue;

		if(mask_str) {
			if(strchr(mask_str, '.') != '\0') {
				if(!inet_aton(mask_str, &ptr[i].mask)) continue;
			} else {
				ptr[i].mask.s_addr = 0;
#ifndef BIGENDIAN
				for (x = 0; x < atoi(mask_str); x++)
					ptr[i].mask.s_addr |= 1 << x;
#else
				for (x = 0; x < atoi(mask_str); x++)
					ptr[i].mask.s_addr |= 1 << (31 - x);
#endif
			}		
		} else
			inet_aton("255.255.255.255", &ptr[i].mask);
	}
}


int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen) {
	int retval;

	if (!initialized)
		if (!initialize()) exit(-1); /* fatal error */

	if ( (((struct sockaddr_in *)my_addr)->sin_addr.s_addr == INADDR_ANY) && (((struct sockaddr_in *)my_addr)->sin_family == AF_INET) )
		memcpy(&((struct sockaddr_in *)my_addr)->sin_addr, &bind_addr, sizeof(struct in_addr));
		
	retval = ((int(*)(int a, const struct sockaddr *b, int c)) bind_p)(sockfd, my_addr, addrlen);        

	if (retval) 
		perror("socketlock: error in bind()");
	return retval;
}


int connect(int  sockfd,  const  struct sockaddr *serv_addr, socklen_t addrlen) {
        int i, dont_bind = 0;
	static struct sockaddr_in new_addr;

	if (!initialized)
		if (!initialize()) exit(-1); /* fatal error */
					       
	if ( ((struct sockaddr_in *)serv_addr)->sin_family == AF_INET ) 
		for (i = 0; (except_nets[i].addr.s_addr && i < MAX_EXCEPTIONS); i++) {
			if ( (((struct sockaddr_in *)serv_addr)->sin_addr.s_addr & except_nets[i].mask.s_addr) 
			== (except_nets[i].addr.s_addr & except_nets[i].mask.s_addr) )
				dont_bind = 1;
		}

	if (!dont_bind) {
		memcpy(&new_addr.sin_addr, &bind_addr, sizeof(struct in_addr));
		new_addr.sin_family = AF_INET;
		
		if(bind(sockfd, (struct sockaddr*)&new_addr, sizeof(struct sockaddr_in)))
			perror("socketlock: error in bind() at connect()");
	}
	
	return ((int(*)(int a, struct sockaddr *b, int c))connect_p)(sockfd, (struct sockaddr *)serv_addr, addrlen);
}
